library synapse;

Synapse synapse = Synapse();

void signal(var input){
  synapse.signal(input);
}

typedef Future<dynamic> NeuronFunction<E>(E element);

/// In the nervous system, a synapse is a structure that permits a neuron
/// (or nerve cell) to pass an electrical or chemical signal to another neuron
/// or to the target effector cell.
class Synapse {

  Synapse(){
    print("Synapse()");
  }

  List<Neuron> _handlers = [];

  List<NeuronMemory> memories() {
    return _handlers.fold<List<NeuronMemory>>([], (previousValue, element) {
      previousValue.addAll(element.memories);
      return previousValue;
    }).toList();
  }

  void signal<T>([var value]) {
    print("synapse.emit(value: $value)");
    _handlers.forEach((handler) {

      if (!handler.enabled) return;

      if (value != null && !handler.canHandle(value)) return;

      if (value == null && !handler.canHandleType<T>()) return;

      Future futureResult = handler.handle(value).catchError((error) {
        signal(error);
        return null;
      });

      if (futureResult == null) {
        return;
      }

      futureResult.then((output) {
        if (output != null) {
          if (handler.canHandle(output)) {
            handler.enabled = false;
            print("handler disabled because it returns own handle type");
          } else {
            signal(output);
          }
        }
      });
    });
  }

  void handle<T>(NeuronFunction<T> _handlerFunction, {String name}) {
    print("synapse.handle()");
    _handlers.add(Neuron<T>(_handlerFunction, name: name));
  }

  void clear() {
    print("synapse.clear()");
    _handlers.clear();
  }
}

class NeuronMemory {
  final Neuron neuron;
  final dynamic input;
  final dynamic output;
  final DateTime start;
  final DateTime end;
  final Exception error;
  bool resolved;

  NeuronMemory({this.input, this.output, this.error, this.start, this.end, this.neuron}){
    if(error != null){
      resolved = false;
    }
  }
}

/// A neuron, neurone or nerve cell, is an electrically
/// excitable cell that communicates with other cells via specialized
/// connections called synapses.
class Neuron<T> {

  NeuronFunction<T> _function;
  bool enabled = true;
  List<NeuronMemory> memories = [];
  String name;

  Neuron(this._function, {this.name});


  Future handle(var value) async {
    DateTime start = DateTime.now();
    try {
      var output = await _function(value);
      memories.add(
          NeuronMemory(input: value,
              output: output,
              error: null,
              start: start,
              end: DateTime.now(),
              neuron: this)
      );
      return output;
    }catch(error){
      memories.add(
          NeuronMemory(input: value,
              output: null,
              error: error,
              start: start,
              end: DateTime.now(),
              neuron: this));
      throw error;
    }
  }

  bool canHandle(var value){
    return value is T;
  }

  bool canHandleType<J>(){
    return J == T;
  }
}


